package clustering;

import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import xxl.core.functions.Function;

public class KMeans implements KMeansClustering {
    
    static Logger logger=Logger.getLogger("Test");

    final static boolean DEBUG = false;
    
    static public Validity MYVF2 = new Validity() {
        public double invoke(KMeansState state) {
            Distance distance = state.kmc.getDistance();
            double[] minClusterDist = new double[state.clusterCount];
            int[] sumC = new int[state.clusterCount];
            double[] sumD = new double[state.clusterCount];
            for (int i = 0; i < state.dataCount; i++) {
                    int c=state.cluster[i];
                    double dist = state.clusterSquaredDist[i][c];
                    sumD[c] += dist;
                    sumC[c] ++;
            }
            if (state.clusterCount == 1) {
                minClusterDist[0] = 0;
            } else {
                for (int c = 0; c < state.clusterCount; c++) {
                    minClusterDist[c] = 0;
                }
                for (int c = 0; c < state.clusterCount - 1; c++) {
                    for (int c2 = c + 1; c2 < state.clusterCount; c2++) {
                        double distc = distance.squaredDistance(state.kMeans[c], state.kMeans[c2]);
                            minClusterDist[c] += 1/((sumC[c]*distc)/sumD[c]);
                            minClusterDist[c2] += 1/((sumC[c2]*distc)/sumD[c2]);
                    }
                }
            }
            double result = 0;
            double[] valCl = new double[state.clusterCount];
            StringBuffer log = new StringBuffer();
            log.append("" + state.clusterCount + "\n");
            for (int c = 0; c < state.clusterCount; c++) {
                if (sumC[c] < 3)
                    valCl[c] = Double.MAX_VALUE / state.maxClusterCount;
                else if (minClusterDist[c] == 0) {
                    if (state.clusterCount == 1)
                        valCl[c] = 0;
                    else
                        valCl[c] = Double.MAX_VALUE / state.maxClusterCount;
                } else
                    valCl[c] = (sumD[c] * minClusterDist[c]);// / (sumC[c]);
                if (Double.isInfinite(valCl[c])) {
                    valCl[c] = Double.MAX_VALUE / state.maxClusterCount;
                }
                if (state.kMeans[c] instanceof double[]) log.append(((double[]) state.kMeans[c])[0] + "," + ((double[]) state.kMeans[c])[1] + "  ");
                log.append((sumD[c] / sumC[c]) + ":" + minClusterDist[c] + " = " + valCl[c] + "    " + sumD[c] +"    " + sumC[c] + "\n");
                result += valCl[c];
                if (Double.isInfinite(result)) {
                    result = Double.MAX_VALUE;
                }
            }
            //result = result / state.clusterCount;
            if (Double.isInfinite(result)) {
                result = Double.MAX_VALUE;
            }
            state.validity = result;
            logger.log(Level.FINER, log.toString());
            logger.log(Level.FINE, state.clusterCount + "     " + result + "\n\n");
            return result;
        }
    };

    static public Validity LERMAN = new Validity() {
        public double invoke(KMeansState state) {
            KMeans km = (KMeans) state.kmc;
            double separation = 0d;
            double sumDist = 0f;
            double sumDist2 = 0f;
            double n2 = state.dataCount * state.dataCount;
            Object mean = null;
            mean = km.computeMeans(state.data, state.dataCount, null);
            for (int i = 0; i < state.dataCount; i++) {
                double dist = km.distance.squaredDistance(state.data[i], mean);
                sumDist += Math.sqrt(dist);
                sumDist2 += dist;
            }
            // double mid=sumDist*2.0/dataCount;
            double midCenter = sumDist / state.dataCount;
            double std = Math
                    .sqrt(sumDist2 / state.dataCount - midCenter * midCenter);

            double[] sums = new double[state.clusterCount];
            for (int i = 0; i < state.dataCount; i++) {
                double dist = state.clusterSquaredDist[i][state.cluster[i]];
                sums[state.cluster[i]] += dist - midCenter;
            }
            sumDist = 0f;
            sumDist2 = 0f;
            for (int i = 0; i < state.clusterCount; i++) {
                sumDist += 2 * state.clusterCounts[i] * sums[i];
                sumDist2 += state.clusterCounts[i] * state.clusterCounts[i];
            }

            separation = (1.0 / Math.sqrt(sumDist2 * (n2 - sumDist2)))
                    * (1 / std) * sumDist;
            // System.out.println("ClusterCount:"+clusterCount+"
            // Separation:"+separation);
            // System.out.println("FINISHED computeValidity");
            return separation;
        }
    };

    static public Validity SEPARATION = new Validity() {
        public double invoke(KMeansState state) {
            KMeans km = (KMeans) state.kmc;
            double separation = 0d;
            double sumDist = 0f;
            double minSqDist = Double.MAX_VALUE;
            int c = 0;
            for (int i = 0; i < state.clusterCount; i++) {
                for (int j = i + 1; j < state.clusterCount; j++) {
                    double dist = km.distance.distance(state.kMeans[i], state.kMeans[j]);
                    sumDist += dist;
                    c++;
                    if (dist < minSqDist)
                        minSqDist = dist;
                }
            }
            // double[] addFactors1=new double[clusterCount];
            // double[] addFactors2=new double[clusterCount];
            // double[] sum2=new double[clusterCount];
            for (int i = 0; i < state.dataCount; i++) {
                double dist = state.clusterSquaredDist[i][state.cluster[i]];
                // System.out.println(dist);
                separation += dist * dist;
            }
            sumDist /= c;
            // double val2=0.5*(minDist+sumDist);
            // double val2=Math.sqrt(minDist*sumDist);
            double val2 = minSqDist;
            separation = separation / (val2);
            // System.out.println("ClusterCount:"+clusterCount+"
            // MinDist:"+minSqDist+" Separation:"+separation);
            return separation;
        }
    };

    static public Validity MYVF = new Validity() {
        public double invoke(KMeansState state) {
            double[] maxDataDist = new double[state.clusterCount];
            double[] minClusterDist = new double[state.clusterCount];
            double[] sumC = new double[state.clusterCount];
            double sum = 0;
            
            Distance distance=state.kmc.getDistance();
            for (int i = 0; i < state.dataCount; i++) {
                int c = state.cluster[i];
                double dist = state.clusterSquaredDist[i][state.cluster[i]];
                if (dist > maxDataDist[c]) {
                    maxDataDist[c] = dist;
                }
                sumC[c] += 1;
                sum += 1;
            }
            if (state.clusterCount == 1) {
                minClusterDist[0] = 0;
            } else {
                for (int c = 0; c < state.clusterCount; c++) {
                    minClusterDist[c] = Double.MAX_VALUE;
                }
                for (int c = 0; c < state.clusterCount - 1; c++) {
                    for (int c2 = c + 1; c2 < state.clusterCount; c2++) {
                        double distc = distance.squaredDistance(state.kMeans[c],
                                state.kMeans[c2]);
                        if (distc < minClusterDist[c]) {
                            minClusterDist[c] = distc;
                        }
                        if (distc < minClusterDist[c2]) {
                            minClusterDist[c2] = distc;
                        }
                    }
                }
            }
            double cc = 3.0;
            double off = 1.0;
            double result = 0;
            double[] valCl = new double[state.clusterCount];
            for (int c = 0; c < state.clusterCount; c++) {
                double penalty = 0;
                if (sumC[c] >= cc)
                    penalty = 1.0;
                else if (sumC[c] <= off)
                    penalty = 0.0;
                else
                    penalty = (sumC[c] - off) / (cc - off);
                if (minClusterDist[c] + maxDataDist[c] == 0) {
                    if (state.clusterCount == 1)
                        valCl[c] = 0;
                    else
                        valCl[c] = 1;
                } else
                    valCl[c] = maxDataDist[c]
                            / (minClusterDist[c] + maxDataDist[c]);
                valCl[c] = Math.pow(valCl[c], penalty);
                result += valCl[c];
            }
            if (result == 0) {
                System.out.println();
            }
            return (result / state.clusterCount);
        }
    };

    static public Objective OBJECTIVE = new Objective() {
        public double invoke(KMeansState state) {

            return 0.0;
        }
    };

    static public Initialization INITALIZE = new Initialization() {
        public void invoke(
                KMeansState state) {
            
            int sp = state.dataCount / state.clusterCount - 1;
            for (int i = 0; i < state.clusterCount; i++) {
                state.kMeans[i] = state.kmc.getCopyFunction().invoke(state.data[i * sp], state.kMeans[i]);
            }
        }
    };

    protected Distance distance;

    protected Validity validity;

    protected Objective objective;

    protected Initialization initialize;

    protected Function copyObject;

    protected Average meansFunction;

    public KMeans(Distance distance, Objective objective, Validity validity,
            Initialization initialize, Function copyObject, Average meansFunction) {
        this.distance = distance;
        this.validity = validity;
        this.objective = objective;
        this.initialize = initialize;
        this.copyObject = copyObject;
        this.meansFunction = meansFunction;
    }

    public KMeansState invokeKMeans(Object[] data, int dataCount, int clusterCount) {
        
        Object[] kmeans = new Object[clusterCount];
        
        int[] cluster = new int[dataCount];
        double[] clusterCounts = new double[dataCount];
        double[][] clusterDist = new double[dataCount][clusterCount];
        
        KMeansState state= new KMeansState(data,dataCount,clusterCount);
        state.clusterSquaredDist=clusterDist;
        state.clusterCounts=clusterCounts;
        state.cluster=cluster;
        state.maxClusterCount=clusterCount;
        state.minClusterCount=clusterCount;
        state.kmc=this;
        state.kMeans=kmeans;
        
        initialize(state);
        
        computeMemberships(state);
        int dif = 0;
        do {
            computeClusterCenters(state);
            dif = computeMemberships(state);
        } while (dif > 0);
        return state;
    }

    /**
     * Initialize the Clustercenters
     */
    public void initialize(KMeansState state) {
        initialize.invoke(state);
    }

    /**
     * Compute the Center of the data of Object-array.
     * 
     * @param obs
     *            The data to compute the center
     * @param len
     *            count of data
     * @param means
     *            Object to store the center. If null, a new Object has to
     *            generate.
     * @return the center
     */
    protected Object computeMeans(Object[] obs, int len, Object means) {
        return meansFunction.average(this, obs, len, means);
    }

    /**
     * Compute the cluster memberships of all data
     * @return count of changed memberships (only if testOld==true)
     */
    public int computeMemberships(KMeansState state) {
        // System.out.println("computeMemberships");

        int changed = 0;

        for (int i = 0; i < state.clusterCount; i++) {
            state.clusterCounts[i] = 0;
        }
        for (int i = 0; i < state.dataCount; i++) {
            double min = distance.squaredDistance(state.data[i], state.kMeans[0]);
            state.clusterSquaredDist[i][0] = min;
            int minCluster = 0;
            for (int k = 1; k < state.clusterCount; k++) {
                double d = distance.squaredDistance(state.data[i], state.kMeans[k]);
                state.clusterSquaredDist[i][k] = d;
                if (d < min) {
                    min = d;
                    minCluster = k;
                }
            }
            
            // System.out.println(i+" in "+minCluster);
            if (state.cluster[i] != minCluster) changed++;
            state.cluster[i] = minCluster;
            state.clusterCounts[minCluster]++;
        }
        if (DEBUG)
            System.out.println("changed: " + changed);
        return changed;
    }

    /**
     * Compute Centers of each Cluster
     */
    public void computeClusterCenters(KMeansState state) {
        // System.out.println("computeClusterCenters");

        Object[][] lists = new Object[state.clusterCount][state.dataCount];
        for (int i = 0; i < state.clusterCount; i++) {
            state.clusterCounts[i] = 0;
        }
        for (int i = 0; i < state.dataCount; i++) {
            int cl = state.cluster[i];
            try {
                lists[cl][(int)state.clusterCounts[cl]] = state.data[i];
            } catch (ArrayIndexOutOfBoundsException e) {
                e.printStackTrace();
                System.out.println(cl + "  " + i + "  " + state.clusterCounts[cl]
                        + "  " + lists.length + "  " + state.clusterCounts.length
                        + "  " + state.data.length);
                System.exit(1);
            }
            state.clusterCounts[cl]++;
        }
        for (int i = 0; i < state.clusterCount; i++) {
            if (state.clusterCounts[i] == 0) {
            } else {
                state.kMeans[i] = computeMeans(lists[i], (int)state.clusterCounts[i], state.kMeans[i]);
            }
        }
    }

   
    public double computeValidity(KMeansState state) {
        return validity.invoke(state);
    }
    
    public double computeObjective(KMeansState state) {
        return objective.invoke(state);
    }

    public Distance getDistance() {
        return distance;
    }

    public Function getCopyFunction() {
        return copyObject;
    }

}
