/*
 * StreamCluster.java
 *
 * Created on 3. Februar 2003, 11:17
 */


package pipes;
import xxl.core.collections.queues.ArrayQueue;
import xxl.core.pipes.operators.AbstractPipe;
import xxl.core.pipes.operators.identities.BufferPipe;
import xxl.core.util.ObjectPool;

/**
 *
 * @author  Jrgen Beringer
 */
public class NormWeightedStreams extends AbstractPipe{
	
	class Block {
	
		double[] objects;
		double sum;
		double sum2;
		double wsum;
		double wsum2;
	}
	
	static boolean DEBUG=false;
	
	boolean bounded=true;
	
	//Anzahl von Streams
	int size;
	//Grsse eines Blocks
	int blockSize;
	//Anzahl von Blcken
	int blockCount;
	//Gewichtungskonstante
	double weight;
	//ObjectPool fr double-Arrays
	ObjectPool doublePool;
	//Grsse des Gesamten Fensters
	int windowSize;
	//Wurzel der Fenstergrsse (oft bentigt)
	double sqrtWS;
	//Queues fr die Blcke
	ArrayQueue[] blockQueues;
	//Block-Array der aktuellen Blcke
	Block[] blocks;
	//Gewichte fr den letzten Block (oft bentigt, teuer zu berechnen)
	double[] weights;
	double[] allWeights;
	//Anfangsgewicht des 2.Blocks
	double blockweight;
	//Gewicht des ersten Elements nach dem Fenster.
	double lastblockweight;
	//Hifswerte(komplexe Einheitswurzeln) zur Berechnung der DFT-Werte des 1.Blocks
	double[] sum;
	//qadratische Summe aller Strme
	double[] sum2;
	//gewichtete Mittelwerte
	double[] mid;
	double[] lastMid;
	//gewichtete Standartabweichung
	double[] derive;
	double[] lastDerive;
	//Summe aller Gewichte ber das gesamte Fenster
	double sumWeights;
	//Decoupler-Object
	BufferPipe bufferPipe;
	//initialisiert? also ob die ersten Cluster schon berechnet wurden
	boolean initialized;
	//anzahl von verarbeiteten Blcken
	int init;
	//Anzahl von Strmen der Cluster
	int[] clusterCounts;
	//Anzahl von unvernderten Neuklassifizierungen  in Folge
	int zerolength;
	
	//hilfsvariablen
	double[][] values1;
	
	//double[][] values2;
	double[][] valuesf1;
	
	//double[][] values11;
	//double[][] values12;
	//double[] valuesf11;
	
	//Beschriftung(Namen) der Streams
	Object[] streamsObjects;
	
	
	
	/** Creates a new instance of StreamCluster */
	public NormWeightedStreams(Object[] streamsObjects, int size, int blockSize, int blockCount, double weight, ObjectPool doublePool) {
		this.streamsObjects=streamsObjects;
		this.size=size;
		this.blockSize=blockSize;
		this.blockCount=blockCount;
		this.weight=weight;
		this.doublePool=doublePool;
		windowSize=blockSize*blockCount;
		sqrtWS=(double)Math.sqrt(windowSize);
		//dftweights2=new double[windowSize][2];
		sum=new double[size];
		sum2=new double[size];
		mid=new double[size];
		derive=new double[size];
		lastMid=new double[size];
		lastDerive=new double[size];
		computeArcs();
		if (bounded) blockQueues=new ArrayQueue[size];
		blocks=new Block[size];
		for (int i=0;i<size;i++){
			blocks[i]=new Block();
			if (bounded){
				blockQueues[i]=new ArrayQueue();
				for (int j=0;j<blockCount;j++){
					blockQueues[i].enqueue(new Block());
				}
			}
		}
		initialized=false;
		init=0;
		
		values1=new double[size][windowSize];
		//values2=new double[windowSize][2];
		valuesf1=new double[size][windowSize];
		//values11=new double[windowSize][2];
		//values12=new double[windowSize][2];
		//valuesf11=new double[windowSize];
		
	}
	
	
	public void setBufferPipe(BufferPipe bufferPipe){
		this.bufferPipe=bufferPipe;
	}
	
	public synchronized void processObject(Object ob,int id){
        
		if (DEBUG) System.out.println("Block "+(init+1));
        init++;
        long start=System.nanoTime();
		
		Object[] input=(Object[])ob;
		for (int i=0;i<size;i++){
			blocks[i].objects=(double[])input[i];
		}
		
		for (int i=0;i<size;i++){
			for (int p=blockSize;p<windowSize;p++){
				valuesf1[i][p-blockSize]=valuesf1[i][p];
			}
		}
		for (int i=0;i<size;i++){
			for (int p=0;p<blockSize;p++){
				valuesf1[i][p+windowSize-blockSize]=blocks[i].objects[p];
			}
		}
		
		
		//compute zeta for all new blocks, and release Object-Memory
		for (int i=0;i<size;i++){
			computeZeta(blocks[i]);
			//doublePool.releaseObject(blocks[i].objects);
			blocks[i].objects=null;
		}
		
		//compute new dft-Values and new weighted middle and derivation
		for (int i=0;i<size;i++){
			Block newBlock=blocks[i];
			Block last=null;
			last =(Block)blockQueues[i].dequeue();
			
			sum[i]=sum[i]*blockweight;
			sum2[i]=sum2[i]*blockweight;
			sum[i]+=newBlock.wsum;
			sum2[i]+=newBlock.wsum2;
			sum[i]-=last.wsum*lastblockweight;
			sum2[i]-=last.wsum2*lastblockweight;
			mid[i]=sum[i]/sumWeights;
			derive[i]=sum2[i]/sumWeights-mid[i]*mid[i];
			if (derive[i]<=0.0) derive[i]=1.0;
			else derive[i]=Math.sqrt(derive[i]);
			
			//System.out.print(" derive:"+i+" ="+windowSize+","+sum[i]+","+sum2[i]+","+mid[i]+","+derive[i]+"; ");
			blockQueues[i].enqueue(newBlock);
			blocks[i]=last;
			//blockPool.releaseObject(last);
		}
		
		
		for (int p=0;p<windowSize;p++){
			for (int i=0;i<size;i++){
				values1[i][p]=((valuesf1[i][p]-mid[i])/derive[i])*allWeights[p];
			}
		}
		
		
		double[]help=lastMid;
		lastMid=mid;
		mid=help;
		
		help=lastDerive;
		lastDerive=derive;
		derive=help;
		/*System.out.print("Norm  ");
		for (int i=0;i<size;i++){
			System.out.print(lastDerive[i]+"  ");
		}
		System.out.println();
		*/
		transfer(values1);
        long end=System.nanoTime();
        //System.out.println(init+"    "+((end-start)/1000000.0));
	}
	
	public double getStdDerive(int nr){
		return lastDerive[nr];
	}
	
	public double getMiddle(int nr){
		return lastMid[nr];
	}
	
	
	
	//berechnet exp(j*2*pi*(b-i)*m)
	public void computeArcs(){
		
		weights=new double[blockSize];
		weights[blockSize-1]=1.0f;
		for (int i=1;i<blockSize;i++){
			weights[blockSize-1-i]=(double)Math.pow(weight,i);
		}
		allWeights=new double[windowSize];
		allWeights[windowSize-1]=1.0f;
		for (int i=1;i<windowSize;i++){
			allWeights[windowSize-1-i]=(double)Math.pow(weight,i);
		}
		lastblockweight=(double)Math.pow(weight,blockSize*(blockCount));
		blockweight=(double)Math.pow(weight,blockSize);
		System.out.println("weigth:"+weight+" windowsize:"+(blockSize*(blockCount))+" lastblockweight:"+lastblockweight);
		sumWeights=0.0;
		double w;
		double[] helpWeights=new double[windowSize];
		for (int p=0;p<windowSize;p++){
			w=Math.pow(weight,(windowSize-1-p));
			sumWeights+=w;
			/*dftweights2[p][0]=w;
			dftweights2[p][1]=0.0;*/
			helpWeights[p]=(double)w;
		}
		
		//fft(dftweights2);
	}
	
	//berechnet zeta-Zahl fr einen Block
	public void computeZeta(Block block){
		
		double sum=0.0;
		double sum2=0.0;
		double wsum=0.0;
		double wsum2=0.0;
		double[] objects=block.objects;
		for (int i=0;i<blockSize;i++){
			sum+=objects[i];
			sum2+=objects[i]*objects[i];
			wsum+=objects[i]*weights[i];
			wsum2+=objects[i]*objects[i]*weights[i];
		}
		
		block.sum=sum;
		block.sum2=sum2;
		block.wsum=wsum;
		block.wsum2=wsum2;
	}
	
}

