/*
 * Created on 18.03.2003
 *
 * To change this generated comment go to 
 * Window>Preferences>Java>Code Generation>Code Template
 */
package pipes;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Panel;
import java.awt.ScrollPane;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;

import xxl.core.collections.queues.ListQueue;
import xxl.core.pipes.operators.identities.BufferPipe;
import xxl.core.pipes.scheduler.RoundRobinStrategy;
import xxl.core.pipes.scheduler.StrategyProcessor;
import xxl.core.pipes.sinks.AbstractSink;
import xxl.core.pipes.sources.Source;

/**
 * @author beringer
 */
public class StreamVisualizer extends AbstractSink {

    protected Image back;

    protected Graphics graphics;

    protected int count;

    protected int height;

    protected int width;

    protected int length;

    protected double max;

    protected Font font;

    protected Object[] streamsObjects;

    public double[][] values;

    protected Frame frame;

    protected Panel panel;

    protected boolean disposed = false;

    protected boolean begin = true;

    protected int number;

    static boolean DEBUG = false;

    protected static final int BLOCKWIDTH = 20;

    protected static final int BLOCKOFFSET = 5;

    protected static int NAMEWIDTH = 110;

    protected Color[] colors = new Color[] { Color.red, new Color(0, 255, 0),
            Color.blue, Color.yellow, Color.magenta, Color.cyan, Color.orange,
            Color.pink, new Color(127, 127, 0), new Color(127, 0, 127),
            new Color(127, 63, 63), new Color(0, 127, 127),
            new Color(127, 127, 255), new Color(127, 255, 0),
            new Color(255, 127, 0), new Color(255, 127, 127),
            new Color(127, 255, 127), new Color(127, 0, 255),
            new Color(255, 0, 127), new Color(127, 0, 0), new Color(0, 127, 0),
            new Color(0, 0, 127), new Color(255, 127, 255),
            new Color(63, 196, 255), new Color(255, 255, 127),
            new Color(127, 196, 63), Color.gray, Color.black };

    public StreamVisualizer(Object[] streamsObjects, int length, int width,
            int height, double max) {
        this(streamsObjects, length, streamsObjects.length, width, height, max,
                null);
    }

    public StreamVisualizer(Object[] streamsObjects, int length, int count,
            int width, int height, double max, Font font) {
        this.count = count;
        this.length = length;
        this.height = height;
        this.width = width;
        this.max = max;
        this.font = font;
        this.streamsObjects = streamsObjects;
        number = 0;
        back = new BufferedImage(width + NAMEWIDTH + BLOCKOFFSET,
                (height + BLOCKOFFSET) * count + 2 * BLOCKOFFSET,
                BufferedImage.TYPE_INT_ARGB);
        graphics = back.getGraphics();
        drawNames(graphics);
        values = new double[count][length];
        final int theHeight = back.getHeight(null);
        final int theWidth = back.getWidth(null);
        panel = new Panel() {

            public void update(Graphics g) {
                paint(g);
            }

            public void paint(Graphics g) {
                if (DEBUG)
                    System.out.println("repaint");
                g.drawImage(back, 0, 0, frame);
            };

            public Dimension getPreferredSize() {
                return new Dimension(theWidth, theHeight);
            }

        };

        panel.setSize(theWidth, theHeight);
    }

    public Component getComponent() {
        return panel;
    }

    public ScrollPane getScrollPane() {
        ScrollPane sp = new ScrollPane();
        sp.add(panel);
        return sp;
    }

    public Frame createFrame(String name) {
        frame = new Frame(name);
        frame.setSize(back.getWidth(frame) + 40, Math.min(500, back
                .getHeight(frame) + 40));
        frame.addWindowListener(new WindowAdapter() {

            public void windowClosing(WindowEvent e) {
                frame.setVisible(false);
                stop();
                disposed = true;
                frame.dispose();
            }
        });

        frame.setLayout(new BorderLayout());
        frame.add(getScrollPane(), "Center");
        frame.setVisible(true);
        return frame;
    }

    public void setBounds(int x, int y, int width, int height) {
        frame.setBounds(x, y, width, height);
    }

    public void drawNames(Graphics g) {
        if (font != null)
            g.setFont(font);
        g.setColor(Color.black);
        for (int i = 0; i < count; i++) {
            g.drawString(streamsObjects[i].toString(), BLOCKOFFSET, i
                    * (height + BLOCKOFFSET) + BLOCKOFFSET + height / 2);
        }
    }

    public synchronized void drawValues(Graphics g, int id, boolean remove,
            int stream) {
        if (remove) {
            g.setColor(Color.white);
            g.fillRect(NAMEWIDTH + BLOCKOFFSET, id * (height + BLOCKOFFSET)
                    + BLOCKOFFSET, width, height + 1 + BLOCKOFFSET);

            g.setColor(Color.black);
            g.drawLine(NAMEWIDTH + BLOCKOFFSET, id * (height + BLOCKOFFSET)
                    + BLOCKOFFSET + (height / 2), width + NAMEWIDTH
                    + BLOCKOFFSET, id * (height + BLOCKOFFSET) + BLOCKOFFSET
                    + (height / 2));
        }
        int posX, lastX, posY, lastY;
        g.setColor(colors[stream]);
        lastX = NAMEWIDTH + BLOCKOFFSET;
        lastY = (int) (-values[id][0] / max * (height / 2.0) + id
                * (height + BLOCKOFFSET) + BLOCKOFFSET + (height / 2.0));
        for (int i = 1; i < length; i++) {
            posX = (int) (i * (width / (double) length) + NAMEWIDTH + BLOCKOFFSET);
            posY = (int) (-values[id][i] / max * (height / 2.0) + id
                    * (height + BLOCKOFFSET) + BLOCKOFFSET + (height / 2.0));
            g.drawLine(lastX, lastY, posX, posY);
            lastX = posX;
            lastY = posY;
        }
    }

    public synchronized void processObject(Object ob, int id) {
        if (DEBUG)
            System.out.println("Input");
        Object[] newObs = null;
        if (ob instanceof Object[])
            newObs = (Object[]) ob;
        else
            newObs = new Object[] { ob };
        if ((begin) && (number >= length)) {
            begin = false;
        }
        int aktCount = Math.min(count, newObs.length);
        int len = 0;
        for (int i = 0; i < aktCount; i++) {
            double[] newVal = (double[]) newObs[i];
            len = newVal.length;
            /*
             * System.arraycopy(values[i],0,values[i],len,length-len);
             * System.arraycopy(newVal,0,values[i],0,len);
             */
            for (int j = 0; j < length - len; j++) {
                values[i][j] = values[i][j + len];
            }
            for (int j = 0; j < len; j++) {
                values[i][length - len + j] = newVal[j];
            }
            /*
             * double sum=0f; for (int j=0;j<length;j++){
             * sum+=values[i][j]*values[i][j]; }
             * System.out.println("SUM2="+sum);
             */
            if (!begin)
                drawValues(graphics, i, (id == 0), id);
        }
        if (begin)
            number += len;
        panel.repaint();
        if (disposed) {
            new Thread() {

                public void run() {
                    frame.dispose();
                }
            }.start();
        } else if (!begin)
            if (frame != null)
                frame.setVisible(true);

    }

    public void done(int ID) {
        if (frame != null) {
            frame.setVisible(false);
            frame.dispose();
        }
    }

    public void writeStreamData() {
        try {
            PrintStream ps = new PrintStream(new FileOutputStream(
                    "C:\\dataStream.txt"));
            ps.print("Key\t");

            for (int j = 0; j < length; j++) {
                ps.print("V" + j + "\t");
            }
            ps.print("Clusternumber\n");

            for (int j = 0; j < count; j++) {
                ps.print(streamsObjects[j] + "\t");
                for (int i = 0; i < length; i++) {
                    ps.print(values[j][i]);
                    if ((i + 1) < length)
                        ps.print("\t");
                }
                if ((j + 1) < count)
                    ps.print("\n");
            }
            ps.flush();
            ps.close();
        } catch (Exception e) {
        }
    }

    
}