/*
 * Created on 24.04.2003
 *
 * To change this generated comment go to 
 * Window>Preferences>Java>Code Generation>Code Template
 */
package pipes;

import java.io.Writer;
import java.util.logging.Level;
import java.util.logging.Logger;

import xxl.core.functions.Function;
import xxl.core.pipes.operators.AbstractPipe;
import xxl.core.pipes.operators.identities.BufferPipe;
import clustering.FuzzyKMeansState;
import clustering.KMeansClustering;
import clustering.KMeansState;
import clustering.RelativeKMeans;

/**
 * @author beringer
 */
public abstract class StreamingKMeans extends
        AbstractPipe<Object[], KMeansState> {

    static protected Logger log = Logger.getLogger("Test");

    protected double lambda = 400;

    protected KMeansClustering kmAlg;

    protected RelativeKMeans rkm;

    protected KMeansState state;

    // Decoupler-Object
    protected BufferPipe bufferPipe;

    protected int initCount;

    // initialisiert? also ob die ersten Cluster schon berechnet wurden
    protected boolean initialized;

    // anzahl von verarbeiteten Blcken
    protected int init;

    protected int changed = 1;

    protected int lastClusterCount;

    protected boolean maxTest;

    protected double[] quality;

    // Beschriftung(Namen) der Streams
    protected String[] streamNames;

    public Writer clusterFile = null;

    public Writer qualityFile = null;

    public int[][] comparer = null;

    public int[] comparerCount = null;

    public boolean normer = false;

    public Function rerun = null;

    public StreamingKMeans(RelativeKMeans rkm, String[] streamNames,
            int size, int blockCount, int clusterCount) {
        this(rkm, streamNames, size, blockCount, clusterCount, clusterCount,
                clusterCount);
    }

    public StreamingKMeans(RelativeKMeans rkm, String[] streamNames,
            int size, int blockCount, int maxClusterCount, int clusterCount) {
        this(rkm, streamNames, size, blockCount, maxClusterCount, 2,
                clusterCount);
    }

    /** Creates a new instance of StreamCluster */
    public StreamingKMeans(RelativeKMeans rkm, String[] streamNames,
            int size, int blockCount, int maxClusterCount, int minClusterCount,
            int clusterCount) {

        this.kmAlg = rkm.getKMeansClustering();
        this.rkm = rkm;
        this.streamNames = streamNames;
        this.initCount = blockCount;

        state = new FuzzyKMeansState();
        state.kmc = kmAlg;
        state.dataCount = size;
        state.maxClusterCount = maxClusterCount;
        state.minClusterCount = minClusterCount;
        state.clusterCount = clusterCount;
        state.kMeans = new Object[maxClusterCount];
        state.cluster = new int[size];
        state.clusterSquaredDist = new double[size][maxClusterCount];
        state.clusterCounts = new double[maxClusterCount];

        initialized = false;
        init = 0;
    }

    public void setBufferPipe(BufferPipe decoupler) {
        this.bufferPipe = decoupler;
    }

    public void processObject(Object[] normdfts, int id) {
        log.log(Level.FINER, "Block " + (init + 1));

        state.data = normdfts;
        init++;
        log.log(Level.FINE, "Step " + init);

        if (initialized) {

            computeClusterCenters();
            computeMemberships();
            computeClusterCenters();
            computeMemberships();
            computeClusterCenters();
            computeMemberships();
            computeClusterCenters();

            rkm.testClusterCount(state);

            iteration();

            computeClusterCenters();
            transfer(state);
            log.log(Level.FINER,"sep:" + state.validity);
            // if ((init+1)%10==0) showClusters();
            // output2();
        } else {
            if (init == initCount + 1) {
                // int sp=size/clusterCount;
                kmAlg.initialize(state);

                maxTest = false;
                initialized = true;
                kmAlg.computeMemberships(state);

                log.log(Level.FINE, "Result:1");

                computeClusterCenters();
                computeMemberships();
                computeClusterCenters();
                computeMemberships();
                computeClusterCenters();

                log.log(Level.FINER, "initialized");

                int oldClusterCount;
                int iter = 0;
                do {
                    oldClusterCount = state.clusterCount;
                    rkm.testClusterCount(state);
                    iter++;
                } while (oldClusterCount != state.clusterCount && iter < 20);
                iteration();
                computeClusterCenters();
                transfer(state);
            }
        }
    }

    protected void computeMemberships() {
        if (maxTest)
            kmAlg.computeMemberships(state);
        changed = kmAlg.computeMemberships(state);
    }

    protected void computeClusterCenters() {
        kmAlg.computeClusterCenters(state);
        if (maxTest)
            kmAlg.computeClusterCenters(state);
    }

    protected void iteration() {
        int c = 0;
        do {
            computeMemberships();
            computeClusterCenters();
            c++;
            // System.out.println("Iteration:"+c+"
            // "+decoupler.getBuffer().size());
        } while ((bufferPipe.getBufferSize() == 0) && (c < 10) && (changed > 0));
        log.log(Level.FINER, "its: " + c + "    " + state.clusterCount);
    }

    public Object[] getClusterMeans() {
        return state.kMeans;
    }

    public int getCluster(int stream) {
        return state.cluster[stream];
    }

    public int getClusterCount() {
        return state.clusterCount;
    }

    public int getMaxClusterCount() {
        return state.maxClusterCount;
    }

}