package applications.pipes;

import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Panel;
import java.awt.ScrollPane;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Random;

import javax.swing.JFrame;
import javax.swing.JSplitPane;

import pipes.Blocker2;
import pipes.ClusterVisualizer;
import pipes.DFTNormWeightedStreams;
import pipes.FuzzyStreamCluster;
import pipes.MultiSource;
import pipes.StreamVisualizer;
import pipes.data.SyntheticRealIterators;
import pipes.data.SyntheticRealIterators.FactoryFunction;
import xxl.core.collections.queues.ArrayQueue;
import xxl.core.math.functions.RealFunction;
import xxl.core.pipes.operators.identities.BufferPipe;
import xxl.core.pipes.scheduler.Controllable;
import xxl.core.pipes.scheduler.RoundRobinStrategy;
import xxl.core.pipes.scheduler.StrategyProcessor;
import xxl.core.pipes.sources.Source;

public class FuzzyChange implements Runnable {

    public static void main(String[] args) {
        
        final JFrame fr = new JFrame("Control");

        fr.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                fr.setVisible(false);
                fr.dispose();
                System.exit(0);
            }
        });
        
        int x = Toolkit.getDefaultToolkit().getScreenSize().width;
        int y = Toolkit.getDefaultToolkit().getScreenSize().height;

                
        
        // Logger logger=Logger.getLogger("Test");
        // logger.setLevel(Level.FINE);
        // Handler handler=new StreamHandler(System.err,new SimpleFormatter());
        // handler.setLevel(Level.FINE);
        // logger.addHandler(handler);

        // Test with 2 Clusters and 5 Streams which change the clusters smoothly
        //test(3,fr.getContentPane(),x,y-60);
        


        // Test with 2 Clusters which split smoothly to 4 clusters and vice
        // versa
        testAll(fr.getContentPane(),x,y-60);
        
        fr.setSize(x, y - 60);
        fr.setVisible(true);
    }

    // weight for changing data streams
    public static double lastW = 0;

    // Test with 2 Clusters and ch Streams which change the clusters smoothly
    public static void test(int ch, final Container con, final int width,final int height) {
        int cl = 3;
        int count = 2 * cl + ch;
        final int blockSize = 256;

        int blockCount = 16;
        int dftCount = 25;
        int maxClasses = 4;
        double lastWeight = 0.4;
        double weight = Math.pow(lastWeight, 1.0 / (blockSize * blockCount));

        int windowSize = blockSize * blockCount;
        String[] obsStreams = new String[count];
        int[] lengths = new int[count];
        boolean[] autosize = new boolean[count];
        double[] min = new double[count];
        double[] max = new double[count];
        for (int i = 0; i < count; i++) {
            obsStreams[i] = "Stream Nr." + i;
            lengths[i] = windowSize;
            autosize[i] = true;
            min[i] = 0;
            max[i] = 1;
        }

        Random random = new Random(227452743);
        FactoryFunction generator = new FactoryFunction(count, blockSize,
                blockCount, random, 2);

         final RealFunction
         func1=SyntheticRealIterators.sinus(blockSize*blockCount/3,60,0);
         final RealFunction
         func2=SyntheticRealIterators.sinus(blockSize*blockCount/1.5,60,Math.PI/4);



        final RealFunction funcCh = SyntheticRealIterators.linear(blockSize
                * blockCount, blockSize * blockCount / 8, 0, 1);

        for (int i = 0; i < cl; i++) {
            generator.exchangeFunction(i, func1,0);
            generator.exchangeFunction(i + cl + ch, func2,0);
        }
        for (int i = 0; i < ch; i++) {
            final int val = i;
            RealFunction func3 = new RealFunction() {
                final int val2 = val;

                public double eval(double val) {
                    double w = funcCh.eval(val);
                    // if (lastW<0)
                    lastW = w;
                    return func1.eval(val) * w + func2.eval(val) * (1 - w);
                }
            };
            generator.exchangeFunction(i + cl, func3,0);
        }

        final MultiSource ms = new MultiSource(generator, count, 750, blockSize);
        ms.setCloseMode(Source.AUTOMATIC_CLOSE);

        Blocker2 bl = new Blocker2(count, 1);
        for (int i = 0; i < count; i++) {
            ms.subscribe(bl, i, i);
            bl.addSource(ms, i);
        }
        BufferPipe dc = new BufferPipe(bl, 2, new ArrayQueue(), 1, 1);
        final StrategyProcessor proc = new StrategyProcessor(new RoundRobinStrategy(
                new Controllable[] { dc }), 0, 2500);
        proc.start();

        DFTNormWeightedStreams dft = new DFTNormWeightedStreams(obsStreams,
                count, blockSize, blockCount, dftCount, weight, bl
                        .getFloatPool());
        dft.addSource(dc, 0);
        dc.subscribe(dft, 0);

        
        // Visualize data streams
        StreamVisualizer sv3 = new StreamVisualizer(obsStreams, blockSize
                * blockCount, width-350, (height-60)/count, 70);
        sv3.addSource(dc, 0);
        dc.subscribe(sv3, 0);
        // JScrollPane com3= new JScrollPane(sv3.getComponent());
        ScrollPane com3 = sv3.getScrollPane();

        FuzzyStreamCluster wcs =
        // new FuzzyStreamCluster(obsStreams, count, blockCount, dftCount,
        // maxClasses, 6, 2.0);
        new FuzzyStreamCluster(obsStreams, count, blockCount, dftCount,
                maxClasses, 2, 3, 2.0, 0.0);
        dft.subscribe(wcs, 0);
        wcs.addSource(dft, 0);

        wcs.setBufferPipe(dc);

        
        ClusterVisualizer cv = new ClusterVisualizer(obsStreams, maxClasses,
                true);
        wcs.subscribe(cv, 0);
        cv.addSource(wcs, 0);
        // JScrollPane comcv= new JScrollPane(cv.getComponent());
        ScrollPane comcv = cv.getScrollPane();
        // sv.start();

        final Panel controlPanel = new Panel();

        Button but2 = new Button("PAUSE");
        but2.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                ms.getProcessor().hasToPauseUntilWakeUp();
                proc.hasToPauseUntilWakeUp();
            }
        });
        Button but3 = new Button("RESUME");
        but3.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                proc.wakeUp();
                ms.getProcessor().wakeUp();
            }
        });
        controlPanel.setLayout(new FlowLayout());
        controlPanel.add(but2);
        controlPanel.add(but3);

        Panel pan = new Panel(){
            public Dimension getPreferredSize() {
                return new Dimension(220,height);
            }  
        };
        pan.setLayout(new BorderLayout());
        pan.add("Center", comcv);
        pan.add("South", controlPanel);
        pan.setSize(width/3,height);

        JSplitPane sp1 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, true, pan,
                com3){
            public Dimension getPreferredSize() {
                return con.getSize();
            }   
        };
        sp1.setDividerLocation(0.33);
        sp1.setBorder(null);

        con.removeAll();
        con.setLayout(new FlowLayout());
        con.add(sp1);
        sp1.setDividerLocation(0.33);
        com3.doLayout();
        ms.open();

    }

    // Test with 2 Clusters which split smoothly to 4 clusters and vice versa
    public static void testAll(Container con, final int width,final int height) {
        int cl = 5;
        int count = 4 * cl;
        final int blockSize = 128;

        int blockCount = 8;
        int dftCount = 10;
        int maxClasses = 4;
        double lastWeight = 0.4;
        double weight = Math.pow(lastWeight, 1.0 / (blockSize * blockCount));

        int windowSize = blockSize * blockCount;
        String[] obsStreams = new String[count];
        int[] lengths = new int[count];
        boolean[] autosize = new boolean[count];
        double[] min = new double[count];
        double[] max = new double[count];
        for (int i = 0; i < count; i++) {
            obsStreams[i] = "Stream Nr." + i;
            lengths[i] = windowSize;
            autosize[i] = true;
            min[i] = 0;
            max[i] = 1;
        }

        Random random = new Random(227452747);
        FactoryFunction generator = new FactoryFunction(2, count, blockSize,
                random);

        final RealFunction func1 = SyntheticRealIterators.sinus(blockSize
                * blockCount / 4, 40, 0);
        final RealFunction func2 = SyntheticRealIterators.sinus(blockSize
                * blockCount / 4, 40, Math.PI);
        final RealFunction func3 = SyntheticRealIterators.sinus(blockSize
                * blockCount / 4, 40, Math.PI / 2);
        final RealFunction func4 = SyntheticRealIterators.sinus(blockSize
                * blockCount / 4, 40, 3 * Math.PI / 2);
        
        final RealFunction funcCh = SyntheticRealIterators.linear(blockSize
                * blockCount, blockSize * blockCount * 2, 0, 1);

        for (int i = 0; i < cl; i++) {
            generator.exchangeFunction(i * 2, func1);
            generator.exchangeFunction(i * 2 + 1, func2);
            RealFunction func5 = new RealFunction() {
                public double eval(double val) {
                    double w = funcCh.eval(val);
                    // if (lastW<0)
                    lastW = w;
                    return func1.eval(val) * w + func3.eval(val) * (1 - w);
                }
            };
            generator.exchangeFunction(i * 2 + 2 * cl, func5);
            RealFunction func6 = new RealFunction() {
                public double eval(double val) {
                    double w = funcCh.eval(val);
                    // if (lastW<0)
                    // lastW=w;
                    return func2.eval(val) * w + func4.eval(val) * (1 - w);
                }
            };
            generator.exchangeFunction(i * 2 + 1 + 2 * cl, func6);
        }

        final MultiSource ms = new MultiSource(generator, count, 1500, blockSize);
        ms.setCloseMode(Source.AUTOMATIC_CLOSE);

        // VisualizeStreams visStreams = new VisualizeStreams("Streams", count,
        // obsStreams, lengths, autosize, min, max, 0, null, null);

        Blocker2 bl = new Blocker2(count, 1);
        for (int i = 0; i < count; i++) {
            ms.subscribe(bl, i, i);
            bl.addSource(ms, i);

            // ms.subscribe(visStreams, i);
            // visStreams.addSource(ms, i);
        }
        BufferPipe dc = new BufferPipe(bl, 2, new ArrayQueue(), 1, 1);
        StrategyProcessor proc = new StrategyProcessor(new RoundRobinStrategy(
                new Controllable[] { dc }), 0, 2500);
        proc.start();

        DFTNormWeightedStreams dft = new DFTNormWeightedStreams(obsStreams,
                count, blockSize, blockCount, dftCount, weight, bl
                        .getFloatPool());
        dft.addSource(dc, 0);
        dc.subscribe(dft, 0);

        int l2 = 0;

        if (width > 1024) {
            l2 = 750;
        } else if (width > 800) {
            l2 = 550;
        } else {
            l2 = 330;
        }

//      Visualize data streams
        StreamVisualizer sv3 = new StreamVisualizer(obsStreams, blockSize
                * blockCount, l2, 70, 120);
        sv3.addSource(dc, 0);
        dc.subscribe(sv3, 0);
        // JScrollPane com3= new JScrollPane(sv3.getComponent());
        ScrollPane com3 = sv3.getScrollPane();

        FuzzyStreamCluster wcs =
        // new FuzzyStreamCluster(obsStreams, count, blockCount, dftCount,
        // maxClasses, 6, 2.0);
        new FuzzyStreamCluster(obsStreams, count, blockCount, dftCount,
                maxClasses, 2, 2, 2.0, 0.0);
        dft.subscribe(wcs, 0);
        wcs.addSource(dft, 0);

        wcs.setBufferPipe(dc);

//        SpecialOutput so = new SpecialOutput(new int[] { 0, 1, 2 * cl,
//                2 * cl + 1 }, 2 * cl);
//        wcs.subscribe(so, 0);
//        so.addSource(wcs, 0);
        /*
         * ClusterTimeVisualizer ctv=new
         * ClusterTimeVisualizer(obsStreams,classes,100); wcs.subscribe(ctv, 0);
         * ctv.addSource(wcs,0);
         */
        ClusterVisualizer cv = new ClusterVisualizer(obsStreams, maxClasses,
                true);
        wcs.subscribe(cv, 0);
        cv.addSource(wcs, 0);
        // JScrollPane comcv= new JScrollPane(cv.getComponent());
        ScrollPane comcv = cv.getScrollPane();
        // sv.start();

        final Panel controlPanel = new Panel();

        Button but2 = new Button("PAUSE");
        but2.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                ms.getProcessor().hasToPauseUntilWakeUp();
            }
        });
        Button but3 = new Button("RESUME");
        but3.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                ms.getProcessor().wakeUp();
            }
        });
        controlPanel.setLayout(new FlowLayout());
        controlPanel.add(but2);
        controlPanel.add(but3);

        Panel pan = new Panel(){
            public Dimension getPreferredSize() {
                return new Dimension(250,height);
            }  
        };
        pan.setLayout(new BorderLayout());
        pan.add("Center", comcv);
        pan.add("South", controlPanel);

        JSplitPane sp1 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, true, pan,
                com3);
        sp1.setDividerLocation(0.3);
        sp1.setBorder(null);

        sp1.setMinimumSize(new Dimension(20, 20));
        sp1.setOneTouchExpandable(false);
        sp1.setContinuousLayout(false);
        
        con.add(sp1);
        
        sp1.setDividerLocation(0.3);
        com3.doLayout();
        ms.open();

    }
    
    public void run(){
        final JFrame fr = new JFrame("Control");

        fr.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                fr.setVisible(false);
                fr.dispose();
                System.exit(0);
            }
        });

        int x = Toolkit.getDefaultToolkit().getScreenSize().width;
        int y = Toolkit.getDefaultToolkit().getScreenSize().height;

        
        test(3,fr.getContentPane(), x, y - 60);

        fr.setSize(x, y - 60);
        fr.setVisible(true);
    }

}
