package applications.pipes;

import java.util.Random;

import pipes.Blocker2;
import pipes.ClusterVisualizer;
import pipes.DFTNormWeightedStreams;
import pipes.FuzzyStreamCluster;
import pipes.MultiSource;
import pipes.data.SyntheticRealIterators;
import xxl.core.pipes.operators.identities.BufferPipe;
import xxl.core.pipes.scheduler.Controllable;
import xxl.core.pipes.scheduler.RoundRobinStrategy;
import xxl.core.pipes.scheduler.StrategyProcessor;
import xxl.core.pipes.sources.Source;

public class FuzzyClusterStreams {

    /**
     * @param args
     */
    public static void main(String[] args) {
        test1();
    }
    
    public static void test1() {
        //Number of Datastreams
        int count = 30;
        //Size of a Block
        int blockSize = 128;
        //Number of Blocks in a Window
        int blockCount = 8;
        //Number of used DFT-Coefficients
        int dftCount = 30;
        //maximal Number of classes
        int maxClasses = 12;
        //weight of the last Element in the Window
        double lastWeight = 0.5;
        //compute the weight value
        double weight = Math.pow(lastWeight, 1.0 / (blockSize * blockCount));
        //Array for the names of the streams
        String[] obsStreams = new String[count];
        
        //Set names of the streams
        for (int i = 0; i < count; i++) {
            obsStreams[i] = "Stream Nr." + i;
        }

        //Get synthetic data streams
        final MultiSource ms = new MultiSource(SyntheticRealIterators
                .getFactory(6, count, blockSize,new Random()), count, 1000,
                blockSize);
        //Close source automatically
        ms.setCloseMode(Source.AUTOMATIC_CLOSE);
        
        //Block streams to process Algorithms in Blocks
        Blocker2 bl = new Blocker2(count, 1);
        //Add Sources for blocking
        for (int i = 0; i < count; i++) {
            ms.subscribe(bl, i,i);
            bl.addSource(ms, i);
        }
        //Add Bufferpipe to buffer the streams and run Algorithm with in extra Thread
        BufferPipe dc = new BufferPipe(bl, 1, 1);
        StrategyProcessor proc = new StrategyProcessor(new RoundRobinStrategy(
                new Controllable[] { dc }), 0, 2500);
        proc.start();

        //Add Normalization to the streams
        DFTNormWeightedStreams dft = new DFTNormWeightedStreams(obsStreams,
                count, blockSize, blockCount, dftCount, weight, bl
                        .getFloatPool());
        dft.addSource(dc, 0);
        dc.subscribe(dft, 0);


        //Add Algorithm-Pipe for Clustering the streams 
        FuzzyStreamCluster wcs = new FuzzyStreamCluster(obsStreams, count,
                blockCount, dftCount, maxClasses,2, 6, 2.0);
        dft.subscribe(wcs, 0);
        wcs.addSource(dft, 0);

        wcs.setBufferPipe(dc);

        //Visualize the Results with a specific Sink
        ClusterVisualizer cv = new ClusterVisualizer(obsStreams, maxClasses,
                true);
        wcs.subscribe(cv, 0);
        //create output Frame
        cv.createFrame("Cluster");
        cv.addSource(wcs, 0);

        
        //Open SOURCE to start the production of the data
        ms.open();
    }

}
